' GREP.BAS UNIX-like grep command based on WALKDIRS.BAS
'
' Usage:
'   grep [-f] [-n] "PATTERN" PATHNAMES
'   -f        fast string matching
'   -n        include line numbers in the output
'   < FILE    redirect input from the specified FILE
'   > FILE    redirect output to the specified FILE
'   >> FILE   append output to the specified FILE
'   PATTERN   regex composed of . [] ? * + ^ $
'   PATHNAMES list of dir/file paths to search, using * and ? wildcards
'
' Regex syntax:
'   .     matches any character
'   [abc] matches a, b or c
'   \.    matches ., likewise \? matches ?, \\ matches \ and so on
'   ?     matches zero or one of the previous character, reluctantly
'   *     matches zero or more of the previous character, reluctantly
'   +     matches one or more of the previous character, reluctantly
'   ^     matches begin of text
'   $     matches end of text

Command

Sub Command.Init
  regex$ = ""
  Font 1
End Sub

Function Command.Options$
  Command.Options$ = "fn"
End Function

Sub Command.Usage
  Print "Usage: grep [-f] [-n] " Chr$(34) "PATTERN" Chr$(34) " PATHNAMES"
  Print "  -f        fast string matching"
  Print "  -n        include line numbers in the output"
  Print "  < FILE    redirect input from the specified FILE"
  Print "  > FILE    redirect output to the specified FILE"
  Print "  >> FILE   append output to the specified FILE"
  Print "  PATTERN   regex composed of . [] ? * + ^ $"
  Print "  PATHNAMES list of dir/file paths to display, using * and ? wildcards"
End Sub

Sub Command.String(arg$)
  Command.Arg arg$
End Sub

Sub Command.Arg(arg$)
  If regex$ = "" Then
    regex$ = arg$
  Else
    Command.NormalizePathname arg$
    Command.Grep "", arg$
  EndIf
End Sub

Sub Command.File(path$, file$)
  ' no recursive dir walk needed for non-globbed args
End Sub

Sub Command.Dir(path$, dir$)
  ' no recursive dir walk needed for non-globbed args
End Sub

Sub Command.NotFound(path$)
  ' no recursive dir walk needed for non-globbed args
End Sub

Sub Command.DoFile(path$, file$)
  If Asc(file$) = Asc(".") And Not (flags And 1) Then Exit Sub
  Command.Grep path$, file$
End Sub

Sub Command.DoDir(path$, dir$)
  If Not outfile Then Print "Skipping directory " path$
End Sub

Sub Command.Done
  If cmdline$ = "" Then Command.Grep "", "-"
End Sub

Sub Command.Grep(path$, file$)
  Local a$, i, n, n$, fd
  If file$ = "-" Then
    fd = infile
  Else
    Open file$ For input As #3
    If MM.Errno Then Print "No such file: " file$: Exit Sub
    fd = 3
  EndIf
  n = 1
  Do While Not Eof(#fd)
    Line Input #fd, a$
    If flags And 1 Then
      If Instr(a$, regex$) Then
        If flags And 2 Then n$ = Str$(n) Else n$ = ""
        Print #outfile, path$ file$ ":" n$ Tab(Pos + 8 - (Pos-1) Mod 8) a$
      EndIf
    Else
      If Match(regex$, a$) Then
        If flags And 2 Then n$ = Str$(n) Else n$ = ""
        Print #outfile, path$ file$ ":" n$ Tab(Pos + 8 - (Pos-1) Mod 8) a$
      EndIf
    EndIf
    n = n + 1
  Loop
  If fd = 3 Then Close #3
End Sub

Sub Command.NormalizePathname(file$)
  Local i
  i = 1
  Do
    i = Instr(i, file$, "/")
    If Not i Then Exit
    file$ = Left$(file$, i - 1) + Chr$(92) + Mid$(file$, i + 1)
    i = i + 1
  Loop
End Sub

'-------------------------------------------------------------------------------
'
'   Rob Pike's match algorithm, extended with operators [] ? +
'
'-------------------------------------------------------------------------------

Function Match(regex$, text$)
  Local s, n, m, f
  n = Len(regex$)
  m = Len(text$)
  Match = 0
  If Mid$(regex$, 1, 1) = "^" Then
    MatchHere regex$, text$, 2, 1, n, m, f
    Match = f
  Else
    For s = 1 To m
      MatchHere regex$, text$, 1, (s), n, m, f
      If f Then Match = s: Exit For
    Next
  EndIf
End Function

Sub MatchHere(regex$, text$, r, s, n, m, f)
  Local i
  Do While r <= n
    If Mid$(regex$, r) = "$" Then f = s > m: Exit Sub
    If Mid$(regex$, r, 1) = "[" Then
      i = Instr(r + 2, regex$, "]")
      If i Then i = i + 1 Else i = r + 1
    ElseIf Mid$(regex$, r, 1) = Chr$(92) Then
      i = r + 2
    Else
      i = r + 1
    EndIf
    If Mid$(regex$, i, 1) = "*" Then
      MatchStar regex$, text$, r, i, s, n, m, f
      Exit Sub
    EndIf
    If Mid$(regex$, i, 1) = "?" Then
      MatchOptional regex$, text$, r, i, s, n, m, f
      Exit Sub
    EndIf
    If s > m Then f = 0: Exit Sub
    i = r
    MatchChar regex$, text$, r, s, f
    If Not f Then Exit Sub
    If Mid$(regex$, r + 1, 1) = "+" Then
      MatchStar regex$, text$, i, r + 1, s + 1, n, m, f
      Exit Sub
    EndIf
    r = r + 1
    s = s + 1
  Loop
  f = 1
End Sub

Sub MatchStar(regex$, text$, r, i, s, n, m, f)
  Do
    MatchHere regex$, text$, i + 1, (s), n, m, f
    If f Then Exit Sub
    If s > m Then Exit Sub
    MatchChar regex$, text$, (r), s, f
    If Not f Then Exit Sub
    s = s + 1
  Loop
End Sub

Sub MatchOptional(regex$, text$, r, i, s, n, m, f)
  MatchHere regex$, text$, i + 1, (s), n, m, f
  If f Then Exit Sub
  If s > m Then Exit Sub
  MatchChar regex$, text$, (r), s, f
  If Not f Then Exit Sub
  MatchHere regex$, text$, i + 1, s + 1, n, m, f
End Sub

Sub MatchChar(regex$, text$, r, s, f)
  Local i, j
  f = Mid$(regex$, r, 1) = "."
  If f Then Exit Sub
  If Mid$(regex$, r, 1) = "[" Then
    i = Instr(r + 2, regex$, "]")
    If i Then
      j = Instr(r + 1, regex$, Mid$(text$, s, 1))
      f = j > 0 And j < i
      r = i
      Exit Sub
    EndIf
  ElseIf Mid$(regex$, r, 1) = Chr$(92) Then
    r = r + 1
  EndIf
  f = Mid$(regex$, r, 1) = Mid$(text$, s, 1)
End Sub

'-------------------------------------------------------------------------------
'
'   WALKDIRS.BAS
'
'-------------------------------------------------------------------------------

Sub Command()

  Option error continue

  ' command line to parse

  cmdline$ = MM.CmdLine$

  ' initialize

  Command.Init

  ' parse command line options and IO redirects

  flags$ = Command.Options$()

  flags = 0
  infile$ = ""
  appfile$ = ""
  outfile$ = ""

  arg = 0

  Do

    cur = arg
    arg$ = Strtok$(arg, cmdline$, " ")

    If arg$ = "" Then Exit

    If (Asc(arg$) = 45 Or Asc(arg$) = 47) And Mid$(arg$, 2, 1) <> " " Then
      For i = 2 To Len(arg$)
        flag = Instr(flags$, Mid$(arg$, i, 1))
        If Not flag Then Command.Usage: End
        flags = flags Or 2^(flag-1)
      Next i
    ElseIf Asc(arg$) = 60 Then
      infile$ = Mid$(arg$, 2)
      If Not Asc(infile$) Then infile$ = Strtok$(arg, cmdline$, " ")
    ElseIf Left$(arg$, 2) = ">>" Then
      appfile$ = Mid$(arg$, 3)
      If Not Asc(appfile$) Then appfile$ = Strtok$(arg, cmdline$, " ")
    ElseIf Asc(arg$) = 62 Then
      outfile$ = Mid$(arg$, 2)
      If Not Asc(outfile$) Then outfile$ = Strtok$(arg, cmdline$, " ")
    Else
      cur = arg
    EndIf

    If cur < arg Then
      cmdline$ = Left$(cmdline$, cur) + Mid$(cmdline$, arg + 1)
      arg = cur
    EndIf

  Loop

  ' set up IO redirects

  infile = 0
  outfile = 0

  If infile$ <> "" Then
    Open infile$ For input As #1
    If MM.Errno Then Error "Cannot open file " + infile$ + " for reading"
    infile = 1
  EndIf

  If outfile$ <> "" Or appfile$ <> "" Then
    If appfile$ <> "" Then
      Open appfile$ For append As #2
      If MM.Errno Then Error "Cannot open file " + appfile$ + " for writing"
    Else
      Open outfile$ For output As #2
      If MM.Errno Then Error "Cannot open file " + outfile$ + " for writing"
    EndIf
    outfile = 2
  EndIf

  ' walk the directories, expand globs

  arg = 0
  halt = 0 ' setting halt = 1 in a command routine will halt directory walk
  dest$ = Strtok$(arg, cmdline$, " ") ' next argument after arg$

  Do

    arg$ = dest$

    If arg$ = "" Then Exit

    dest$ = Strtok$(arg, cmdline$, " ")

    If Asc(arg$) = 34 Then
      Command.String Mid$(arg$, 2, Len(arg$) - 2)
    ElseIf Instr(arg$, "*") Then
      WalkDirs arg$
    Else
      Command.Arg arg$
    EndIf

  Loop Until halt

  ' finalize

  Command.Done

  If outfile Then Close #outfile
  If infile Then Close #infile

  End

End Sub

' returns next token from str$ at start, separated by sep$

Function Strtok$(start, str$, sep$)
  Local sep
  Do
    start = start + 1
    sep = Instr(start, str$, sep$)
  Loop Until sep <> start
  If Not sep Then
    sep = Len(str$) + 1
    If start >= sep Then Strtok$ = "": Exit Function
  EndIf
  Strtok$ = Mid$(str$, start, sep - start)
  If Asc(Strtok$) = 34 Then
    sep = Instr(start + 1, str$, Chr$(34))
    If Not sep Then sep = Len(str$)
    Strtok$ = Mid$(str$, start, sep + 1 - start)
  EndIf
  start = sep
End Function

' walk the directory tree starting at pathname name$, optional start and path$

Sub WalkDirs(name$, start, path$)
  Local n, i, dir$, file$
  If start <= 0 Then
    start = 1
    If Mid$(name$, 2, 1) = ":" Then
      path$ = Mid$(name$, start, 2)
      Drive path$
      start = 3
    EndIf
  EndIf
  n = Instr(start, name$, Chr$(92))
  If Not n Then n = Instr(start, name$, "/")
  If Not n Then
    If Not Instr(start, name$, "*") Then
      file$ = Dir$(Mid$(name$, start), FILE)
      If file$ = "" Then
        dir$ = Dir$(Mid$(name$, start), DIR)
        If dir$ = "" Then
          Command.NotFound name$
        Else
          Command.Dir path$, dir$
        EndIf
      Else
        Command.File path$, file$
      EndIf
    ElseIf Not Instr(start, name$, ".") Then
      dir$ = Dir$(Mid$(name$, start), DIR)
      Do While dir$ <> "" And Not halt
        Command.DoDir path$, dir$
        dir$ = Dir$()
      Loop
      file$ = Dir$(Mid$(name$ + ".*", start), FILE)
      Do While file$ <> "" And Not halt
        Command.DoFile path$, file$
        file$ = Dir$()
      Loop
    Else
      file$ = Dir$(Mid$(name$, start), FILE)
      Do While file$ <> "" And Not halt
        Command.DoFile path$, file$
        file$ = Dir$()
      Loop
    EndIf
  ElseIf n = start Then
    dir$ = Cwd$
    Chdir Chr$(92)
    WalkDirs name$, n + 1, path$ + Chr$(92)
    Chdir dir$
  Else
    dir$ = Dir$(Mid$(name$, start, n - start), DIR)
    Do While dir$ <> "" And Not halt
      Chdir dir$
      WalkDirs name$, n + 1, path$ + dir$ + Chr$(92)
      Chdir ".."
      dir$ = Dir$()
    Loop
  EndIf
End Sub
