' MORE.BAS UNIX-like more command based on WALKDIRS.BAS
'
' Usage:
'   more [-a] [-n] [-p] [-r] [-x] PATHNAMES
'   -a        show contents of hidden files
'   -n        add line numbers
'   -p        do not pause
'   -r        show raw control characters
'   -x        show hex dump
'   < FILE    redirect input from the specified FILE
'   > FILE    redirect output to the specified FILE
'   >> FILE   append output to the specified FILE
'   PATHNAMES list of dir/file paths to display, using * and ? wildcards

Command

Sub Command.Init
  Font 1
End Sub

Function Command.Options$
  Command.Options$ = "anprx"
End Function

Sub Command.Usage
  Print "Usage: more [-a] [-n] [-p] [-r] [-x] PATHNAMES"
  Print "  -a        show contents of hidden files"
  Print "  -n        add line numbers"
  Print "  -p        do not pause"
  Print "  -r        show raw control characters"
  Print "  -x        show hex dump"
  Print "  < FILE    redirect input from the specified FILE"
  Print "  > FILE    redirect output to the specified FILE"
  Print "  >> FILE   append output to the specified FILE"
  Print "  PATHNAMES list of dir/file paths to display, using * and ? wildcards"
End Sub

Sub Command.String(arg$)
  Command.Arg arg$
End Sub

Sub Command.Arg(arg$)
  Command.NormalizePathname arg$
  Command.More "", arg$
End Sub

Sub Command.File(path$, file$)
  ' no recursive dir walk needed for non-globbed args
End Sub

Sub Command.Dir(path$, dir$)
  ' no recursive dir walk needed for non-globbed args
End Sub

Sub Command.NotFound(path$)
  ' no recursive dir walk needed for non-globbed args
End Sub

Sub Command.DoFile(path$, file$)
  If Asc(file$) = Asc(".") And Not (flags And 1) Then Exit Sub
  Command.More path$, file$
End Sub

Sub Command.DoDir(path$, dir$)
  If Not outfile Then Print "Skipping directory " path$
End Sub

Sub Command.Done
  If cmdline$ = "" Then Command.More "", "-"
End Sub

Sub Command.More(path$, file$)
  Local a$, b, w, h, i, j, k, n, m, x$, fd
  w = MM.HRes\6
  h = MM.VRes\12
  i = 1
  n = 0
  If file$ = "-" Then
    fd = infile
  Else
    Open file$ For input As #3
    If MM.Errno Then Print "No such file: " file$: Exit Sub
    fd = 3
  EndIf
  If flags And 16 Then
    If w <= 40 Then
      m = 8
    Else
      m = 16
    EndIf
    Do While Not Eof(#fd)
      x$ = ""
      a$ = ""
      For j = 1 To m
        b = Asc(Input$(1, #fd))
        If Not fd And b < 127 Then
          If b = 10 Then Print Chr$(13);
          Print Chr$(b);
        EndIf
        If b < 16 Then
          If Not fd And b = 4 Then Exit For
          x$ = x$ + "0" + Hex$(b) + " "
          a$ = a$ + "."
        Else
          x$ = x$ + Hex$(b) + " "
          If b < 32 Or b > 126 Then b = 46
          a$ = a$ + Chr$(b)
        EndIf
        b = 0
        If Eof(#fd) Then Exit For
      Next
      For j = j + 1 To m
        x$ = x$ + "   "
      Next
      x$ = x$ + a$
      a$ = Hex$(n)
      For j = Len(a$) To 6
        a$ = "0" + a$
      Next
      Print #outfile, a$ Tab(9) x$
      If b = 4 Then Exit
      n = n + m
      i = i + 1
      If Not outfile And (Eof(#fd) Or (i = h And Not (flags And 4))) Then
        Font 1,, 1
        Print "more " path$ file$;
        If Eof(#fd) Then Print " (END)";
        Font 1,, 0
        Do: k = Asc(Inkey$): Loop Until k
        Print Chr$(13) Tab(w) Chr$(13);
        If k = 27 Or k = 81 Or k = 113 Then b = 4: Exit
        If k = 13 Then
          i = i - 1
        Else
          i = 1
        EndIf
      EndIf
    Loop
  Else
    If flags And 2 Then w = w - 8
    n = 1
    Do While Not Eof(#fd)
      Do While Not Eof(#fd)
        a$ = ""
        Do While Not Eof(#fd)
          b = Asc(Input$(1, #fd))
          If Not fd And b < 127 Then
            If b = 10 Then Print Chr$(13);
            Print Chr$(b);
          EndIf
          If b = 10 Then Exit
          If b = 13 Then
            b = Asc(Input$(1, #fd))
            If b = 10 Then Exit
            a$ = a$ + "\13" + Chr$(b)
          ElseIf b < 32 Or b > 126 Then
            If Not fd And a$ = "" And b = 4 Then Exit
            If flags And 8 Then
              a$ = a$ + Chr$(b)
            Else
              a$ = a$ + Chr$(92) + Str$(b)
            EndIf
            b = 0
          Else
            a$ = a$ + Chr$(b)
          EndIf
          If Len(a$) > 251 Then Exit
        Loop
        If b = 4 Then Exit
        For j = 0 To Len(a$)\w
          If flags And 2 Then Print Format$(n, "%7g ");
          Print #outfile, Mid$(a$, j*w + 1, w)
          i = i + 1
          If Not outfile And (Eof(#fd) Or (i = h And Not (flags And 4))) Then
            Font 1,, 1
            Print "more " path$ file$;
            If Eof(#fd) Then Print " (END)";
            Font 1,, 0
            Do: k = Asc(Inkey$): Loop Until k
            Print Chr$(13) Tab(w) Chr$(13);
            If k = 27 Or k = 81 Or k = 113 Then
              b = 4
              Exit
            EndIf
            If k = 13 Then
              i = i - 1
            Else
              i = 1
            EndIf
          EndIf
        Next
        If b = 10 Then n = n + 1
      Loop
      If b = 4 Then Exit
    Loop
  EndIf
  If fd = 3 Then Close #3
End Sub

Sub Command.NormalizePathname(file$)
  Local i
  i = 1
  Do
    i = Instr(i, file$, "/")
    If Not i Then Exit
    file$ = Left$(file$, i - 1) + Chr$(92) + Mid$(file$, i + 1)
    i = i + 1
  Loop
End Sub

'-------------------------------------------------------------------------------
'
'   WALKDIRS.BAS
'
'-------------------------------------------------------------------------------

Sub Command()

  Option error continue

  ' command line to parse

  cmdline$ = MM.CmdLine$

  ' initialize

  Command.Init

  ' parse command line options and IO redirects

  flags$ = Command.Options$()

  flags = 0
  infile$ = ""
  appfile$ = ""
  outfile$ = ""

  arg = 0

  Do

    cur = arg
    arg$ = Strtok$(arg, cmdline$, " ")

    If arg$ = "" Then Exit

    If (Asc(arg$) = 45 Or Asc(arg$) = 47) And Mid$(arg$, 2, 1) <> " " Then
      For i = 2 To Len(arg$)
        flag = Instr(flags$, Mid$(arg$, i, 1))
        If Not flag Then Command.Usage: End
        flags = flags Or 2^(flag-1)
      Next i
    ElseIf Asc(arg$) = 60 Then
      infile$ = Mid$(arg$, 2)
      If Not Asc(infile$) Then infile$ = Strtok$(arg, cmdline$, " ")
    ElseIf Left$(arg$, 2) = ">>" Then
      appfile$ = Mid$(arg$, 3)
      If Not Asc(appfile$) Then appfile$ = Strtok$(arg, cmdline$, " ")
    ElseIf Asc(arg$) = 62 Then
      outfile$ = Mid$(arg$, 2)
      If Not Asc(outfile$) Then outfile$ = Strtok$(arg, cmdline$, " ")
    Else
      cur = arg
    EndIf

    If cur < arg Then
      cmdline$ = Left$(cmdline$, cur) + Mid$(cmdline$, arg + 1)
      arg = cur
    EndIf

  Loop

  ' set up IO redirects

  infile = 0
  outfile = 0

  If infile$ <> "" Then
    Open infile$ For input As #1
    If MM.Errno Then Error "Cannot open file " + infile$ + " for reading"
    infile = 1
  EndIf

  If outfile$ <> "" Or appfile$ <> "" Then
    If appfile$ <> "" Then
      Open appfile$ For append As #2
      If MM.Errno Then Error "Cannot open file " + appfile$ + " for writing"
    Else
      Open outfile$ For output As #2
      If MM.Errno Then Error "Cannot open file " + outfile$ + " for writing"
    EndIf
    outfile = 2
  EndIf

  ' walk the directories, expand globs

  arg = 0
  halt = 0 ' setting halt = 1 in a command routine will halt directory walk
  dest$ = Strtok$(arg, cmdline$, " ") ' next argument after arg$

  Do

    arg$ = dest$

    If arg$ = "" Then Exit

    dest$ = Strtok$(arg, cmdline$, " ")

    If Asc(arg$) = 34 Then
      Command.String Mid$(arg$, 2, Len(arg$) - 2)
    ElseIf Instr(arg$, "*") Then
      WalkDirs arg$
    Else
      Command.Arg arg$
    EndIf

  Loop Until halt

  ' finalize

  Command.Done

  If outfile Then Close #outfile
  If infile Then Close #infile

  End

End Sub

' returns next token from str$ at start, separated by sep$

Function Strtok$(start, str$, sep$)
  Local sep
  Do
    start = start + 1
    sep = Instr(start, str$, sep$)
  Loop Until sep <> start
  If Not sep Then
    sep = Len(str$) + 1
    If start >= sep Then Strtok$ = "": Exit Function
  EndIf
  Strtok$ = Mid$(str$, start, sep - start)
  If Asc(Strtok$) = 34 Then
    sep = Instr(start + 1, str$, Chr$(34))
    If Not sep Then sep = Len(str$)
    Strtok$ = Mid$(str$, start, sep + 1 - start)
  EndIf
  start = sep
End Function

' walk the directory tree starting at pathname name$, optional start and path$

Sub WalkDirs(name$, start, path$)
  Local n, i, dir$, file$
  If start <= 0 Then
    start = 1
    If Mid$(name$, 2, 1) = ":" Then
      path$ = Mid$(name$, start, 2)
      Drive path$
      start = 3
    EndIf
  EndIf
  n = Instr(start, name$, Chr$(92))
  If Not n Then n = Instr(start, name$, "/")
  If Not n Then
    If Not Instr(start, name$, "*") Then
      file$ = Dir$(Mid$(name$, start), FILE)
      If file$ = "" Then
        dir$ = Dir$(Mid$(name$, start), DIR)
        If dir$ = "" Then
          Command.NotFound name$
        Else
          Command.Dir path$, dir$
        EndIf
      Else
        Command.File path$, file$
      EndIf
    ElseIf Not Instr(start, name$, ".") Then
      dir$ = Dir$(Mid$(name$, start), DIR)
      Do While dir$ <> "" And Not halt
        Command.DoDir path$, dir$
        dir$ = Dir$()
      Loop
      file$ = Dir$(Mid$(name$ + ".*", start), FILE)
      Do While file$ <> "" And Not halt
        Command.DoFile path$, file$
        file$ = Dir$()
      Loop
    Else
      file$ = Dir$(Mid$(name$, start), FILE)
      Do While file$ <> "" And Not halt
        Command.DoFile path$, file$
        file$ = Dir$()
      Loop
    EndIf
  ElseIf n = start Then
    dir$ = Cwd$
    Chdir Chr$(92)
    WalkDirs name$, n + 1, path$ + Chr$(92)
    Chdir dir$
  Else
    dir$ = Dir$(Mid$(name$, start, n - start), DIR)
    Do While dir$ <> "" And Not halt
      Chdir dir$
      WalkDirs name$, n + 1, path$ + dir$ + Chr$(92)
      Chdir ".."
      dir$ = Dir$()
    Loop
  EndIf
End Sub
